/*
 *
 * Copyright (c) 2019 Texas Instruments Incorporated
 *
 * All rights reserved not granted herein.
 *
 * Limited License.
 *
 * Texas Instruments Incorporated grants a world-wide, royalty-free, non-exclusive
 * license under copyrights and patents it now or hereafter owns or controls to make,
 * have made, use, import, offer to sell and sell ("Utilize") this software subject to the
 * terms herein.  With respect to the foregoing patent license, such license is granted
 * solely to the extent that any such patent is necessary to Utilize the software alone.
 * The patent license shall not apply to any combinations which include this software,
 * other than combinations with devices manufactured by or for TI ("TI Devices").
 * No hardware patent is licensed hereunder.
 *
 * Redistributions must preserve existing copyright notices and reproduce this license
 * (including the above copyright notice and the disclaimer and (if applicable) source
 * code license limitations below) in the documentation and/or other materials provided
 * with the distribution
 *
 * Redistribution and use in binary form, without modification, are permitted provided
 * that the following conditions are met:
 *
 * *       No reverse engineering, decompilation, or disassembly of this software is
 * permitted with respect to any software provided in binary form.
 *
 * *       any redistribution and use are licensed by TI for use only with TI Devices.
 *
 * *       Nothing shall obligate TI to provide you with source code for the software
 * licensed and provided to you in object code.
 *
 * If software source code is provided to you, modification and redistribution of the
 * source code are permitted provided that the following conditions are met:
 *
 * *       any redistribution and use of the source code, including any resulting derivative
 * works, are licensed by TI for use only with TI Devices.
 *
 * *       any redistribution and use of any object code compiled from the source code
 * and any resulting derivative works, are licensed by TI for use only with TI Devices.
 *
 * Neither the name of Texas Instruments Incorporated nor the names of its suppliers
 *
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * DISCLAIMER.
 *
 * THIS SOFTWARE IS PROVIDED BY TI AND TI'S LICENSORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TI AND TI'S LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <utils/console_io/include/app_log.h>
#include <utils/remote_service/include/app_remote_service.h>
#include <utils/mem/include/app_mem.h>
#include <utils/ipc/include/app_ipc.h>

/* #define APP_REMOTE_SERVICE_TEST_DEBUG */
#define MY_REMOTE_SERVICE_TEST_NAME  "com.my.remote_service_test"

#define APP_REMOTE_SERVICE_TEST_CMD_0   (0x1234)
#define APP_REMOTE_SERVICE_TEST_CMD_1   (0x5678)

int32_t myRemoteServiceTestRunCmd0(uint32_t cpu_id)
{
    int32_t status = 0;
    volatile uint32_t value = 0;
    uint32_t i, count = 1000;
    uint32_t test_pass = 1;
    uint64_t cur_time;

    appLogPrintf("REMOTE_SERVICE_TEST: Running test for CPU %s !!!\n",
        appIpcGetCpuName(cpu_id)
        );

    cur_time = appLogGetTimeInUsec();

    for(i=0; i<count; i++)
    {
        status = appRemoteServiceRun(cpu_id, MY_REMOTE_SERVICE_TEST_NAME, APP_REMOTE_SERVICE_TEST_CMD_0, (void*)&value, sizeof(uint32_t), 0);
        if(status!=0 || value != (i+1))
        {
            test_pass = 0;
            status = -1;
            break;
        }
    }

    cur_time = appLogGetTimeInUsec() - cur_time;

    if(test_pass)
    {
        appLogPrintf("REMOTE_SERVICE_TEST: Test passed, execution time = %d usecs (%d usecs per iteration) !!!\n", cur_time, cur_time/count);
    }
    else
    {
        appLogPrintf("REMOTE_SERVICE_TEST: Test failed @ iteration %d !!!\n", i);
    }
    return status;
}

int32_t myRemoteServiceTestRunCmd1(uint32_t cpu_id)
{
    int32_t status = 0;
    uint32_t value = 0;
    uint32_t i, count = 1000;
    uint32_t test_pass = 1;
    uint32_t mem_size = 1024;
    uint64_t phys_ptr;
    void *virt_ptr;
    uint64_t cur_time;

    virt_ptr = appMemAlloc(APP_MEM_HEAP_DDR, mem_size, 32);
    if(NULL == virt_ptr)
    {
        appLogPrintf("REMOTE_SERVICE_TEST: Test failed!!!\n");
        return (-1);
    }

    phys_ptr = appMemGetVirt2PhyBufPtr((uint64_t)(uintptr_t)virt_ptr, APP_MEM_HEAP_DDR);

    /* assume 32b physical pointer */
    value = (uint32_t)phys_ptr;

    appLogPrintf("REMOTE_SERVICE_TEST: Running test @ 0x%08x of %d bytes size for CPU %s !!!\n",
        value,
        mem_size,
        appIpcGetCpuName(cpu_id)
        );

    cur_time = appLogGetTimeInUsec();

    for(i=0; i<count; i++)
    {
        volatile uint32_t before, after;

        before = i*10;

        appMemCacheInv((void*)virt_ptr, 256);

        *(volatile uint32_t*)virt_ptr = before;

        appMemCacheWbInv((void*)virt_ptr, 256);

        status = appRemoteServiceRun(cpu_id, MY_REMOTE_SERVICE_TEST_NAME, APP_REMOTE_SERVICE_TEST_CMD_1, &value, sizeof(uint32_t), 0);

        appMemCacheInv((void*)virt_ptr, 256);

        after = *(volatile uint32_t*)virt_ptr;

        #ifdef APP_REMOTE_SERVICE_TEST_DEBUG
        appLogPrintf("REMOTE_SERVICE_TEST: 0x%08x = %d\n",
            phys_ptr,
            after
            );
        #endif

        #if 1
        if(after != (before+1))
        {
            test_pass = 0;
            break;
        }
        #endif
    }

    cur_time = appLogGetTimeInUsec() - cur_time;

    appMemFree(APP_MEM_HEAP_DDR, virt_ptr, mem_size);
    if(test_pass)
    {
        appLogPrintf("REMOTE_SERVICE_TEST: Test passed, execution time = %d usecs (%d usecs per iteration) !!!\n", cur_time, cur_time/count);
    }
    else
    {
        appLogPrintf("REMOTE_SERVICE_TEST: Test failed @ iteration %d !!!\n", i);
    }
    return status;
}

int32_t myRemoteServiceTestRun(uint32_t cpu_id)
{
    myRemoteServiceTestRunCmd0(cpu_id);
    myRemoteServiceTestRunCmd1(cpu_id);

    return 0;
}
